from dataclasses import dataclass

import pytest

from serieux import deserialize, serialize
from serieux.auto import Auto, Call
from serieux.exc import SchemaError

from .definitions import Point


class Funky:
    def __init__(self, x: int, y: bool):
        self.marks = funky(x, y)


def funky(x: int, y: bool) -> str:
    return ("!" if y else ".") * x


@dataclass
class HoldsFunk:
    funk: Funky
    more: bool


def test_auto():
    funk = deserialize(Auto[Funky], {"x": 3, "y": True})
    assert funk.marks == "!!!"


def test_auto_from_init():
    funk = deserialize(Auto[Funky], {"x": 3, "y": True})
    assert funk.marks == "!!!"


def test_auto_callable():
    funk = deserialize(Auto[funky], {"x": 3, "y": True})
    assert funk() == "!!!"


def test_call_callable():
    funk = deserialize(Call[funky], {"x": 3, "y": True})
    assert funk == "!!!"


def test_call_on_type():
    with pytest.raises(TypeError, match=r"Call\[...\] should only wrap callables"):
        deserialize(Call[Funky], {"x": 3, "y": True})


def test_auto_not_serializable():
    with pytest.raises(SchemaError, match="does not specify how to serialize"):
        serialize(Auto[Funky], Funky(x=3, y=True))


def test_auto_inherit():
    hfunk = deserialize(Auto[HoldsFunk], {"more": True, "funk": {"x": 3, "y": True}})
    assert hfunk.funk.marks == "!!!"


def test_auto_no_interference():
    pt = serialize(Auto[Point], Point(1, 2))
    assert pt == {"x": 1, "y": 2}



import importlib

from ovld import Exactly, Medley, ovld, recurse

from ..ctx import Context
from ..exc import ValidationError
from ..utils import PRIO_LOW

###################
# Implementations #
###################


def _resolve(ref, ctx):
    if ref is None:
        raise ValidationError("There must be a class reference to resolve.", ctx=ctx)

    if (ncolon := ref.count(":")) == 0:
        raise ValidationError(f"Class reference must include module: '{ref}'", ctx=ctx)
    elif ncolon == 1:
        mod_name, symbol = ref.split(":")
    else:
        raise ValidationError(f"Bad format for class reference: '{ref}'", ctx=ctx)
    try:
        mod = importlib.import_module(mod_name)
        return getattr(mod, symbol)
    except (ModuleNotFoundError, AttributeError) as exc:
        raise ValidationError(exc=exc, ctx=ctx)


class AutoTagAny(Medley):
    @ovld(priority=PRIO_LOW)
    def serialize(self, t: type[Exactly[object]], obj: object, ctx: Context, /):
        objt = type(obj)
        qn = objt.__qualname__
        if "." in qn:
            raise ValidationError("Only top-level symbols can be serialized", ctx=ctx)
        mod = objt.__module__
        rval = recurse(objt, obj, ctx)
        if not isinstance(rval, dict):
            rval = {"return": rval}
        rval["class"] = f"{mod}:{qn}"
        return rval

    @ovld(priority=PRIO_LOW)
    def deserialize(self, t: type[Exactly[object]], obj: dict, ctx: Context, /):
        obj = dict(obj)
        cls_name = obj.pop("class", None)
        obj = obj.pop("return", obj)
        if cls_name is not None:
            cls_name = recurse(str, cls_name, ctx)
        actual_class = _resolve(cls_name, ctx)
        assert actual_class
        return recurse(actual_class, obj, ctx)


# Not a default feature
__default_features__ = None



from .autotag import AutoTagAny
from .clargs import FromArguments
from .dotted import DottedNotation
from .fromfile import FromFile, FromFileExtra
from .interpol import Interpolation
from .lazy import LazyDeserialization
from .partial import PartialBuilding
from .tagged import TaggedTypes
from .tsubclass import TaggedSubclassFeature

__all__ = [
    "AutoTagAny",
    "DottedNotation",
    "FromArguments",
    "FromFile",
    "FromFileExtra",
    "Interpolation",
    "LazyDeserialization",
    "PartialBuilding",
    "TaggedSubclassFeature",
    "TaggedTypes",
]



import importlib
from collections import deque
from typing import TYPE_CHECKING, Annotated, TypeAlias

from ovld import Medley, call_next, ovld, recurse

from ..ctx import Context
from ..exc import ValidationError
from ..instructions import NewInstruction, T, strip_all
from ..schema import AnnotatedSchema

#############
# Constants #
#############


if TYPE_CHECKING:
    TaggedSubclass: TypeAlias = Annotated[T, None]
else:
    TaggedSubclass: TypeAlias = NewInstruction[T, "TaggedSubclass", 1, False]


###################
# Implementations #
###################


def _resolve(ref, base, ctx):
    if ref is None:
        return base

    if (ncolon := ref.count(":")) == 0:
        mod_name = base.__module__
        symbol = ref
    elif ncolon == 1:
        mod_name, symbol = ref.split(":")
    else:
        raise ValidationError(f"Bad format for class reference: '{ref}'", ctx=ctx)
    try:
        mod = importlib.import_module(mod_name)
        return getattr(mod, symbol)
    except (ModuleNotFoundError, AttributeError) as exc:
        raise ValidationError(exc=exc, ctx=ctx)


class TaggedSubclassFeature(Medley):
    @ovld(priority=10)
    def serialize(self, t: type[TaggedSubclass], obj: object, ctx: Context, /):
        base = t.pushdown()
        if not isinstance(obj, base):
            raise ValidationError(f"'{obj}' is not a subclass of '{base}'", ctx=ctx)
        objt = type(obj)
        qn = objt.__qualname__
        if "." in qn:
            raise ValidationError("Only top-level symbols can be serialized", ctx=ctx)
        mod = objt.__module__
        rval = call_next(objt, obj, ctx)
        rval["class"] = f"{mod}:{qn}"
        return rval

    def deserialize(self, t: type[TaggedSubclass], obj: dict, ctx: Context, /):
        base = t.pushdown()
        obj = dict(obj)
        cls_name = obj.pop("class", None)
        if cls_name is not None:
            cls_name = recurse(str, cls_name, ctx)
        actual_class = _resolve(cls_name, base, ctx)
        if not issubclass(actual_class, base):
            raise ValidationError(f"'{actual_class}' is not a subclass of '{base}'", ctx=ctx)
        return recurse(TaggedSubclass.strip(t[actual_class]), obj, ctx)

    def schema(self, t: type[TaggedSubclass], ctx: Context):
        base = strip_all(t)
        subschemas = []
        base_mod = base.__module__
        queue = deque([base])
        while queue:
            sc = queue.popleft()
            queue.extend(sc.__subclasses__())
            sc_mod = sc.__module__
            sc_name = sc.__name__
            subsch = recurse(TaggedSubclass.strip(t[sc]))
            subsch = AnnotatedSchema(
                parent=subsch,
                properties={
                    "class": {
                        "description": "Reference to the class to instantiate",
                        "const": sc_name if sc_mod == base_mod else f"{sc_mod}:{sc_name}",
                    }
                },
                required=[] if sc is base else ["class"],
            )
            subschemas.append(subsch)
        return {"oneOf": subschemas}


# Add as a default feature in serieux.Serieux
__default_features__ = TaggedSubclassFeature



from datetime import date, datetime, timedelta
from pathlib import Path
from zoneinfo import ZoneInfo

import pytest

from serieux import deserialize, serialize
from serieux.exc import ValidationError
from serieux.features.fromfile import WorkingDirectory
from serieux.tell import TypeTell, tells

from .test_schema import schema

# Tests for non-principal types


#############
# Test date #
#############


def test_serialize_date():
    assert serialize(date, date(2023, 5, 15)) == "2023-05-15"
    assert serialize(list[date], [date(2023, 5, 15), date(2024, 1, 1)]) == [
        "2023-05-15",
        "2024-01-01",
    ]


def test_serialize_date_error():
    with pytest.raises(ValidationError, match="Cannot serialize"):
        serialize(date, "2023-05-15")
    with pytest.raises(ValidationError, match="Cannot serialize"):
        serialize(list[date], ["2023-05-15", "2024-01-01"])


def test_deserialize_date():
    assert deserialize(date, "2023-05-15") == date(2023, 5, 15)
    assert deserialize(list[date], ["2023-05-15", "2024-01-01"]) == [
        date(2023, 5, 15),
        date(2024, 1, 1),
    ]


def test_schema_date():
    assert schema(date) == {"type": "string", "format": "date"}


def test_tells_date():
    assert tells(date) == {TypeTell(str)}


#################
# Test datetime #
#################


def test_serialize_datetime():
    assert serialize(datetime, datetime(2023, 5, 15, 12, 30, 45)) == "2023-05-15T12:30:45"
    assert serialize(
        list[datetime], [datetime(2023, 5, 15, 12, 30, 45), datetime(2024, 1, 1, 0, 0, 0)]
    ) == ["2023-05-15T12:30:45", "2024-01-01T00:00:00"]


def test_deserialize_datetime():
    assert deserialize(datetime, "2023-05-15T12:30:45") == datetime(2023, 5, 15, 12, 30, 45)
    assert deserialize(list[datetime], ["2023-05-15T12:30:45", "2024-01-01T00:00:00"]) == [
        datetime(2023, 5, 15, 12, 30, 45),
        datetime(2024, 1, 1, 0, 0, 0),
    ]


def test_deserialize_datetime_timestamp():
    assert deserialize(datetime, 1746471553) == datetime.fromtimestamp(1746471553)


def test_schema_datetime():
    assert schema(datetime) == {"type": "string", "format": "date-time"}


def test_tells_datetime():
    assert tells(datetime) == {TypeTell(str)}


##################
# Test timedelta #
##################


def test_serialize_timedelta():
    assert serialize(timedelta, timedelta(seconds=42)) == "42s"
    assert serialize(timedelta, timedelta(seconds=42, microseconds=500000)) == "42500000us"
    assert serialize(timedelta, timedelta(seconds=-10)) == "-10s"
    assert serialize(list[timedelta], [timedelta(seconds=30), timedelta(days=5)]) == [
        "30s",
        "432000s",
    ]


def test_deserialize_timedelta():
    assert deserialize(timedelta, "42s") == timedelta(seconds=42)
    assert deserialize(timedelta, "42500000us") == timedelta(seconds=42, microseconds=500000)
    assert deserialize(timedelta, "-10s") == timedelta(seconds=-10)
    assert deserialize(timedelta, "6h30m") == timedelta(hours=6, minutes=30)
    assert deserialize(timedelta, "+8d") == timedelta(days=8)
    assert deserialize(timedelta, "4.5d") == timedelta(days=4, seconds=60 * 60 * 12)
    assert deserialize(timedelta, "1d12h") == timedelta(days=1, hours=12)
    assert deserialize(timedelta, "36h") == timedelta(days=1, hours=12)
    assert deserialize(list[timedelta], ["30s", "5d"]) == [
        timedelta(seconds=30),
        timedelta(days=5),
    ]
    with pytest.raises(ValidationError, match="is not a valid timedelta"):
        deserialize(timedelta, "1dd")
    with pytest.raises(ValidationError, match="is not a valid timedelta"):
        deserialize(timedelta, "1d3")
    with pytest.raises(ValidationError, match="is not a valid timedelta"):
        deserialize(timedelta, "1d3x")
    with pytest.raises(ValidationError, match="Could not convert"):
        deserialize(timedelta, "1.5.4d")


def test_schema_timedelta():
    assert schema(timedelta) == {
        "type": "string",
        "pattern": r"^[+-]?([\d.]+[dhms]|[\d.]+ms|[\d.]+us)+$",
    }


def test_tells_timedelta():
    assert tells(timedelta) == {TypeTell(str)}


#############
# Test Path #
#############


def test_serialize_path():
    assert serialize(Path, Path("hello/world.txt")) == "hello/world.txt"
    assert (
        serialize(Path, Path("hello/world.txt"), WorkingDirectory(directory=Path("hello")))
        == "world.txt"
    )


def test_deserialize_path():
    assert deserialize(Path, "hello/world.txt") == Path("hello/world.txt")
    assert deserialize(Path, "world.txt", WorkingDirectory(directory=Path("hello"))) == Path(
        "hello/world.txt"
    )


def test_schema_path():
    assert schema(Path) == {"type": "string"}


def test_tells_path():
    assert tells(Path) == {TypeTell(str)}


#################
# Test ZoneInfo #
#################


def test_serialize_zoneinfo():
    assert serialize(ZoneInfo, ZoneInfo("America/New_York")) == "America/New_York"
    assert serialize(ZoneInfo, ZoneInfo("UTC")) == "UTC"


def test_deserialize_zoneinfo():
    assert deserialize(ZoneInfo, "America/New_York") == ZoneInfo("America/New_York")
    assert deserialize(ZoneInfo, "UTC") == ZoneInfo("UTC")
    with pytest.raises(ValidationError, match="No time zone found"):
        deserialize(ZoneInfo, "Invalid/Timezone")


def test_schema_zoneinfo():
    assert schema(ZoneInfo) == {"type": "string"}


def test_tells_zoneinfo():
    assert tells(ZoneInfo) == {TypeTell(str)}



import importlib
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING

from .auto import Auto
from .ctx import AccessPath, Context, Patch, Patcher
from .exc import SerieuxError, ValidationError, ValidationExceptionGroup
from .features.clargs import CLIDefinition, CommandLineArguments, parse_cli
from .features.dotted import DottedNotation
from .features.fromfile import FromFileExtra, WorkingDirectory
from .features.interpol import Environment
from .features.lazy import DeepLazy, Lazy, LazyProxy
from .features.partial import Partial, Sources
from .features.registered import Registered, StringMapped, singleton
from .features.tagged import Tagged, TaggedUnion
from .features.tsubclass import TaggedSubclass
from .impl import BaseImplementation
from .instructions import InstructionType, NewInstruction
from .model import Extensible, FieldModelizable, Model, Modelizable, StringModelizable
from .schema import RefPolicy, Schema
from .utils import JSON, check_signature
from .version import version as __version__


def _default_features():
    # Collect features declared in the __default_features__ global variables of the features
    # in features/. To see which ones this is on a cloned repo, you can run:
    # $ git grep __default_features__
    here = Path(__file__).parent
    features = []
    for name in (here / "features").glob("*.py"):
        if not name.stem.startswith("_"):
            mod = importlib.import_module(f"{__spec__.name}.features.{name.stem}")
            if feat := getattr(mod, "__default_features__", None):
                features.append(feat)
    features.sort(key=lambda t: -len(t.mro()))
    return features


default_features = _default_features()


if TYPE_CHECKING:  # pragma: no cover
    from typing import TypeVar

    T = TypeVar("T")

    JSON = list["JSON"] | dict[str, "JSON"] | int | str | float | bool | None

    class _MC:
        def __add__(self, other) -> type["Serieux"]: ...

    class Serieux(metaclass=_MC):
        def dump(
            self, t: type[T], obj: object, ctx: Context = None, *, dest: Path = None
        ) -> JSON | None: ...

        def load(self, t: type[T], obj: object, ctx: Context = None) -> T: ...

        def serialize(self, t: type[T], obj: object, ctx: Context = None) -> JSON: ...

        def deserialize(self, t: type[T], obj: object, ctx: Context = None) -> T: ...

        def schema(self, t: type[T], ctx: Context = None) -> Schema[str, "JSON"]: ...

        def __add__(self, other) -> "Serieux": ...

else:

    class Serieux(BaseImplementation, *default_features):
        pass


serieux = Serieux()
serialize = serieux.serialize
deserialize = serieux.deserialize
schema = serieux.schema
load = serieux.load
dump = serieux.dump


def serializer(fn=None, priority=0):
    if fn is None:
        return partial(serializer, priority=priority)

    check_signature(fn, "serializer", ("self", "t: type[T1]", "obj: T2", "ctx: T3>Context"))
    Serieux.serialize.register(fn, priority=priority)
    return fn


def deserializer(fn=None, priority=0):
    if fn is None:
        return partial(deserializer, priority=priority)

    check_signature(fn, "deserializer", ("self", "t: type[T1]", "obj: T2", "ctx: T3>Context"))
    Serieux.deserialize.register(fn, priority=priority)
    return fn


def schema_definition(fn=None, priority=0):
    if fn is None:
        return partial(schema_definition, priority=priority)

    check_signature(fn, "schema definition", ("self", "t: type[T1]", "ctx: T2>Context"))
    Serieux.schema.register(fn, priority=priority)
    return fn


__all__ = [
    "__version__",
    "AccessPath",
    "Auto",
    "BaseImplementation",
    "CLIDefinition",
    "CommandLineArguments",
    "Context",
    "DeepLazy",
    "deserialize",
    "DottedNotation",
    "dump",
    "Environment",
    "Extensible",
    "FromFileExtra",
    "InstructionType",
    "JSON",
    "Lazy",
    "LazyProxy",
    "load",
    "Model",
    "FieldModelizable",
    "Modelizable",
    "NewInstruction",
    "parse_cli",
    "Partial",
    "Patch",
    "Patcher",
    "RefPolicy",
    "Registered",
    "schema",
    "Schema",
    "serialize",
    "serieux",
    "Serieux",
    "SerieuxError",
    "singleton",
    "Sources",
    "StringMapped",
    "StringModelizable",
    "Tagged",
    "TaggedUnion",
    "TaggedSubclass",
    "ValidationError",
    "ValidationExceptionGroup",
    "WorkingDirectory",
]
